(***********************************************************************
 *
 * MiniMath version 1.1
 *
 * Copyright (c) 2015 Michel Deslierres
 * All rights reserved. Tous droits réservés.
 *
 * Read full license in the file license.txt
 * Lire la licence au complet dans le fichier license.txt
 *
 *************************************************************************
 *
 * Changements / Changes
 *
 * - version 1.1
 *
 *     Added hyperbolic functions not found in math.pas.
 *
 *************************************************************************)
 
 
{: Unit that provides some mathematical functions not found in Math.pas }

unit MiniMath;

{$I parser.inc}

interface

uses
  parsertypes;

{$IFDEF UNIT_INFO}
const
  SVersionName = 'MiniMath';
  SVersion = '1.1';
  VersionMajor = 1;
  VersionMinor = 1;
{$ENDIF}

 {: Integer division and modulo, combined div and mod }
procedure IDivMod(const X, Y: float; var Q: Integer; var R: float);

  {: Integer modulo division }
function IMod(const X, Y: float): float;

  {: Integer division }
function IDiv(const X, Y: float): Integer;

function Sinh(const X: float): float;
function Cosh(const X: float): float;
function Tanh(const X: float): float;
function CotH(const X: float): float;
function SecH(const X: float): float;
function CscH(const X: float): float;
function ArcSecH(const X: float): float;
function ArcCotH(const X: float): float;
function ArcCscH(const X: float): float;



implementation

uses
  sysutils;

var
  Epsilon: float;

(***************************************************
Reference:

Division and Modulus for Computer Scientists
DAAN LEIJEN
University of Utrecht
Dept. of Computer Science
PO.Box 80.089, 3508 TB Utrecht
The Netherlands
daan@cs.uu.nl,http://www.cs.uu.nl/~daan/lvm.html
December 3, 2001

/* Euclidean division */
long divE( long D, long d )
{
  long q = D/d;
  long r = D%d;
  if (r < 0) {
    if (d > 0) q = q-1;
          else q = q+1;
  }
  return q;
}

long modE( long D, long d )
{
  long r = D%d;
  if (r < 0) {
    if (d > 0) r = r + d;
          else r = r - d;
 }
 return r;
}

*******************************************************)

{$IFDEF EUCLIDEAN_DIV_MOD}

procedure IDivMod(const X, Y: float; var Q: Integer; var R: float);
begin
  Q := trunc(X/Y);
  R := X - Q*Y;
  if R < 0 then begin
    if Y > 0 then begin
      R := R + Y;
      dec(Q)
    end
    else begin
      R := R - Y;
      Inc(Q);
    end;
  end;
end;

function IMod(const X, Y: float): float;
var
  Q: integer;
begin
  //  DivModE(X,Y,Q, result);
  Q := trunc(X/Y);
  result := X - Q*Y;
  if result < 0 then begin
    if Y > 0 then begin
      result := result + Y;
    end
    else begin
      result := result - Y;
    end;
  end;
end;

function IDiv(const X, Y: float): Integer;
begin
  //DivModE(X,Y,result, R);
  result := trunc(X/Y);
  if X - Result*Y < 0 then begin
    if Y > 0 then begin
      dec(result)
    end
    else begin
      Inc(result);
    end;
  end;
end;

{$ELSE}
{Borland TP / Delphi div mod}

procedure IDivMod(const X, Y: float; var Q: Integer; var R: float);
begin
  Q := trunc(X/Y);
  R := X - Q*Y
end;

function IMod(const X, Y: float): float;
begin
  result := X - trunc(X/Y)*Y;
end;

function IDiv(const X, Y: float): Integer;
begin
  result := trunc(X/Y)
end;

{$ENDIF}

function IsZero(const x: float): Boolean;
begin
  Result := Abs(x) <= Epsilon;
end;

function NearZero(const X, eps: float): boolean;
var
 ep: float;
begin
  if Eps <= 0 then
    ep := Epsilon
  else
    ep := eps;
  result := abs(X) <= Ep;
end;

function SameNumber(const x, y, eps: float): Boolean;
var
  a, b, ep: float;
begin
  if Eps = 0 then
    ep := Epsilon
  else
    ep := eps;
  if (y <> 0) and (x <> 0) then begin
    a := abs(x);
    b := abs(y);
    if a > b then
      a := b;
    ep := a*ep
  end;
  Result := abs(x - y) <= ep;
end;

procedure CalcMachineEpsilon;
var
  One: float;
  T: float;
begin
  One := 1.0;
  Epsilon := One;
  repeat
    Epsilon := 0.5 * Epsilon;
    T := One + Epsilon;
  until One = T;
  Epsilon := 2.0 * Epsilon;
end;

function Sinh(const X: float): float;
begin
  if IsZero(X) then
    Result := 0
  else
    Result := (Exp(X) - Exp(-X)) / 2;
end;

function Cosh(const X: float): float;
begin
  if IsZero(X) then
    Result := 1
  else
    Result := (Exp(X) + Exp(-X)) / 2;
end;

function Tanh(const X: float): float;
begin
  if IsZero(X) then
    Result := 0
  else
    Result := SinH(X) / CosH(X);
end;

function CotH(const X: float): float;
begin
  Result := 1 / TanH(X);
end;

function SecH(const X: float): float;
begin
  Result := 1 / CosH(X);
end;

function CscH(const X: float): float;
begin
  Result := 1 / SinH(X);
end;

function ArcCotH(const X: float): float;
begin
  Result := 0.5 * Ln((X + 1) / (X - 1));
end;

function ArcSecH(const X: float): float;
begin
  Result := Ln((Sqrt(1 - X * X) + 1) / X);
end;

function ArcCscH(const X: float): float;
begin
  Result := Ln(Sqrt(1 + (1 / (X * X))) + (1 / X));
end;

initialization
  CalcMachineEpsilon;
end.
