(***********************************************************************
 *
 * ParserTypes version 1.3
 *
 * Copyright (c) 2016 Michel Deslierres
 * All rights reserved. Tous droits réservés.
 *
 * Read full license in the file license.txt
 * Lire la licence au complet dans le fichier license.txt
 *
 *************************************************************************
 *
 * Changements / Changes
 *
 * - version 1.3
 *
 *     Removed meIdentIsBuiltInName
 *
 * - version 1.2
 *
 *     Modified to handle internal functions by adding type dFloatArray
 *     and new parser errors:
 *       meExpectedSeparator, meExpectedLeftPar, meUnknownIdentifier,
 *       meTooFewParams, meTooManyParams.
 *     Also added ttUnknownIdentifier and ttSeparator to the TTokenType enumeration.
 *  
 *    If NO_LAST_RESULTS is not defined, added ttDollar, ttDblDollar to TTokenType enumeration.
 *
 *    If NO_EXTERNALS is not defined, added the meIdentIsBuiltInName to the parser
 *      error enumeration.
 *
 * - version 1.1
 *
 *     Added ttCaret ('^' and '**'), ttColon (':'), ttPercent ('%s')
 *     and ttComment ('//') to the TMathTokenType enumeration.
 *
 *************************************************************************)

{: Unit that contains global types that are used by other units in
   the project. See ParserConsts.pas for string constants.}
Unit ParserTypes;

{$IFDEF FPC}
  {$mode objfpc}
  {$h+ }
{$ENDIF}  

{$I parser.inc}

interface

uses
  SysUtils;

{$IFDEF UNIT_INFO}
const
  SVersionName = 'ParserTypes';
  SVersion = '1.3';
  VersionMajor = 1;
  VersionMinor = 3;
{$ENDIF}

const
  nan = 0/0;

type
  float = extended;
  dFloatArray = array of float;

type
  TMathTokenType = (

    {: Value returned by NextToken or NextTrueToken when a token starting
       with an unknown character is found (unknown initial characters are
       any characters that are not a space (ASCII code 1..32), not a digit
       not an operator and not ASCII #0 and not a letter.
       In that case TokenLength is 1.}
    ttUnknown,

    {: Value returned by NextToken or NextTrueToken when a token starting
       with a letter ['A'..'Z', 'a'..'z'] is not the name of a targe.}
    ttUnknownIdentifier,

    {: Value returned by NextToken when a sequence of one or more
       consecutive white spaces has been found. This type is never
       returned by NextTrueToken.}
    ttWhite,

    {: Value returned by NextToken or NextTrueToken when a number has
       been found. A number has the following form
         1. - sequence of one or more digits
           1.1 - [optional] decimal symbol (see property DecimalChar)
             1.1.1 - [optional] sequence of digits
         2. [optional] "E" or "e"
           2.1 - [optional] "+" or "-"
           2.2 - sequence of one or more digits.}
    ttNumber,

    {: Value returned by NextToken or NextTrueToken when the end of the
       source has been reached. In that case TokenLength is 0.}
    ttEOS,

    {: Value returned by NextToken or NextTrueToken when the operator
        "-" has been found.}
    ttMinus,

    {: Value returned by NextToken or NextTrueToken when the operator
       "+" has been found.}
    ttPlus,

    {: Value returned by NextToken or NextTrueToken when the operator
       "*" has been found.}
    ttMult,

    {: Value returned by NextToken or NextTrueToken when the operator
       "/" has been found.}
    ttDiv,

    (*: Value returned by NextToken or NextTrueToken when one of "(",
	   "[" or "{" operators has been found.*)
    ttLeftParenthesis,

    (*: Value returned by NextToken or NextTrueToken when one of ")",
	     "]" or "}" operators has been found.*)
    ttRightParenthesis,

    {: Value returned by NextToken or NextTrueToken when the operator
       "^" or "**" has been found.}
    ttCaret,

    {: Value returned by NextToken or NextTrueToken when the operator
       ":" or keyword "div" has been found.}
    ttColon,

    {: Value returned by NextToken or NextTrueToken when the operator
       "%" or keyword "mod" has been found.}
    ttPercent,

    {: Value returned by NextToken or NextTrueToken when the operator
       "//" has been found.}
    ttComment,

    {: Value returned by NextToken or NextTrueToken when an identifier
        which is a word beginning with a letter has been found.}
    ttIdentifier,

    {: Value returned by NextToken or NextTrueToken when the operator
       ListChar ("," on English system, ";" on French systems etc.) has been
       found. The actual character is locale dependant or can be set.}
    ttSeparator

    {$IFnDEF NO_LAST_RESULTS}
    {: Values returned by NextToken or NextTrueToken when the operators
       "$" and "$$" are found.}
    ,ttDollar, ttDblDollar
    {$ENDIF}
    );

type
  {: Errors that the mathematical expression evaluator may report.}
  TMathError = (
    {: Never returned by the parser when exceptions are raised}
    meNoError,

    {: Unexpected end of string when expecting a factor (in module factor).
       This error will be raised when parsing "5 +" because a
       factor should follow the + sign.}
    meUnexpectedEOS,

    {: Expected the end of the expression, found something extraneous.
       This error occurs when parsing the final number in the expression
       "5 + 8 32".}
    meExpectedEOS,

    {: An invalid number was found. This occurs in the Factor routine
       when parsing an expression such as "5,3E*8" where the "E" must
       be followed by a valid positive or negative integer.}
    meInvalidNumber,

    {: Found an unknown or invalid operator when parsing a factor.
       This errors occurs when parsing the multiply operator in the
       expression "5 + *" or when parsing the unknown operator "!"
       in the expression "5 + !".}
    meUnexpectedElement,

    {: Expected a ")" to mark the end of a function call or to mark the
       end of an expression that started with a "(". This error occurs
       in the procedure Builtin when parsing the expressions "sin(5 + 8"
       and "sin(5, 8)"  and in the procedure Factor when parsing a factor
       of the form "( expression )" with a missing right paranthesis as
       in the expression "5*(2 + 3".}
    meExpectedRightPar,
    
    {: Expected list separator (see ParserListChar and SetParserListChar)
       when parsing a function's arguments. This error will occur when
       parsing  "max(4 5)" or "max(4 / 5)".}
    meExpectedSeparator,

    {: Expected a "(" after an function identifier. This error will occur
       when parsing "ln 32".}
    meExpectedLeftPar,

    {: An unknown identifier has been found. Example "lin(32)"
       assuming no built in nor user defined function of that name
       exists.}
    meUnknownIdentifier,

    {: Too few parameters were found when parsing a function's
       parameter list. Example "min()"}
    meTooFewParams,

    {: Too many parameters were found when parsing a function's
       parameter list. Example "sin(32, 34)"}
    meTooManyParams,

    {: Run-time EMathError exception (EInvalidOp, EZeroDivide, EOverflow
       and EUnderflow and EIntError (EDivByZero). }
    meEvalError,

    {: Any other run-time exception including EDivByZero }
    meOtherError);

  TParserError = meUnexpectedEOS..pred(meEValError);

  {: Parser exception }
  EParserError = class(Exception)
  public
    Error: TMathError;
    TokenStart: integer;
    TokenLength: integer;
  end;

const
  ParserErrors: set of TMathError = [low(TParserError)..high(TParserError)];

implementation

end.
