(***********************************************************************
 *
 * Targets 1.1
 *
 * Copyright (c) 2014 Michel Deslierres
 * All rights reserved. Tous droits rservs.
 *
 * Read full license in the file license.txt
 * Lire la licence au complet dans le fichier license.txt
 *
 ************************************************************************
 * - version 1.1
 *
 *     Added extended national language support by storing an internal
 *     encoded name (the en_us name) of each target. The function
 *     InternalName returns the local name of a target given its
 *     encoded name. This is enabled only if the directive NO_NLS2 is
 *     not defined.
 *
 ************************************************************************)

 {: Unit that defines the list of internal (built in) functions and constants.}
unit Targets;

{$IFDEF FPC}
  {$mode objfpc}
  {$h+ }
{$ENDIF}

{$I parser.inc}

interface

uses
   Classes, ParserTypes;

{$IFDEF UNIT_INFO}
const
  SVersionName = 'Targets';
  SVersion = '1.1';
  VersionMajor = 1;
  VersionMinor = 1;
{$ENDIF}

Type
  {: Base class of all named targets. The parser can call on five types of targets:
      - infix keywords mod, div
      - constant (pi, e)
      - variables
      - function with fixed number of arguments (sin, log, ...)
      - function with variable number of arguments (max, min, ...)
   An individual target does not know its own name but it, optionnaly,
   knows its English name.}
  TTarget = class
  private
     FCount: integer;
   protected
     {$IFnDEF NO_TARGET_HINT}
     FHint: string;
     {$ENDIF}
     {$IFnDEF NO_NLS2}
     FInternalName: string;
     {$ENDIF}
     procedure SetCount(value: integer); virtual;
   public
     function Calc: float; virtual; abstract;
     property Count: integer read FCount write SetCount;
     {$IFnDEF NO_TARGET_HINT}
     property Hint: string read FHint write FHint;
     {$ENDIF}
     {$IFnDEF NO_NLS2}
     property InternalName: string read FInternalName write FInternalName;
     {$ENDIF}
   end;

  {: Class of keywords: mod (same as operator %), div (same as operator :).}
  TKeyword = class(TTarget)
  public
    constructor create(anId: integer);
    property Id: integer read FCount;  // TTokenType of operator
    function Calc: float; override;
  end;

  {: Class of built-in constants. Create with the constant's value
     as in TConstant.Create(pi). }
  TConstant = class(TTarget)
  private
    FValue: float;
  public
    constructor create(value: float);
    function Calc: float; override;
  end;

  {$IFnDEF NO_EXTERNALS}
  {: Class of user defined variables defined with assignment 
     statemen (if not added directly by programmer). }
  TVariable = class(TConstant)
  public
    property Value: float read FValue write FValue;
  end;
  {$ENDIF}

  {: Class of named functions with fixed number of parameters.
     Created with number of variables which should be 1 or greater.
     Note that the function does not own the Args array. The call node
     must set the function's ArgArray to its array of args before
     calling Calc.

     Calculation of a built-in function is done in two steps. First
     the array of values, one for each argument of the function
     is calculated by the caller, the built-in function's ArgArray
     is set equal to that array and then the built-in function's
     Calc function is called.

        Func.ArgArray := Args;
        result := Func.Calc

    See Call in the parser.

    The class TFunction does not allocate memory to its FArgs field,
    it does not free any memory to that pointer. This is the
    responsability of the caller. }
  TFunction = class(TTarget)
  public
    Args: dFloatArray;
    constructor create(aCount: integer);
  end;

  {: Class of named functions with variable number of parameters, create with
     the minimum number of expected variables.}
  TVariableFunction = class(TFunction)
  private
    FMinCount: integer;
  protected
    procedure SetCount(value: integer); override;
  public
    constructor create(aMinCount: integer);
    property MinCount: integer read FMinCount;
  end;

  TTargetList = class
  private
    FList: TStrings;
    {$IFDEF DEBUG}{$IFnDEF NO_SIGIL}
    FIsInternalList: boolean;
    {$ENDIF}{$ENDIF}
    function GetCaseSensitive: boolean;
    function GetCount: integer;
    procedure SetCaseSensitive(value: boolean);
    function GetTarget(index: integer): TTarget;
  public
    constructor Create;
    destructor Destroy; override;

    {: Add a target by supplying its name, instantiated class and
       optionaly its Hint. For extended national language support,
       optionaly add the target's encoded name (its en_US name) if 
       it is different from aName. If an error occurs (duplicate 
       name) then target is freed and  an exception is raised. }
    procedure Add(const aName: string; target: TTarget
                  {$IFnDEF NO_TARGET_HINT}; const aHint: string = '' {$ENDIF}
                  {$IFnDEF NO_NLS2}; const eName: string = '' {$ENDIF}
                  ); {$IFnDEF NO_EXTERNALS} virtual; {$ENDIF}

    {: Removes all internal targets from the list }
    procedure Clear;

    {: Removes a named internal target from the list. Returns true
       if aName was found and it's target freed, false otherwise. }
    function Delete(const aName: string): boolean;

    {: Searches for a target by name. If found, returns true and returns
       the found target.}
    function Find(const aName: string; out target: TTarget): boolean;

    {: Returns true if the specified name is found in the list. }
    function HasTarget(const aName: string): boolean;

    {$IFnDEF NO_NLS2}
    {: If the list contains a target with the encoded name eName, returns
       true and the localized name of the target in variable aName.
       Otherwise returns false and aName is empty. }
    function InternalName(const eName: string; out aName: string): boolean; //overload;
    //function LocalName(const eName: string): boolean; overload;
    {$ENDIF}

    {: Returns the name of function, constant or variable of the
        specified target.}
    function NameOf(target: TTarget): string;

    {: If set, then 'Sin' will not be found if the sinus function
       was added to the list as 'sin'. By default, the
       target names are not case sensitive.}
    property CaseSensitive: boolean read GetCaseSensitive write SetCaseSensitive;

    {: Number of targets in list.}
    property Count: integer read GetCount;

    {: Returns the name of all the targets (functions or contants) added
       to the the list.}
    property Names: TStrings read FList;

    {: Access to the targets in alphabetical order }
    property Targets[index: integer]: TTarget read GetTarget; default;
  end;

{$IFnDEF NO_NLS2}

procedure LoadInternalTargets;

{$ENDIF}

var
  {List of all internal built-in functions and constants.}
  InternalTargets: TTargetList;

  {$IFnDEF NO_EXTERNALS}
  {: List of external constants and variables.
     Built-in and user-defined functions are kept in separate lists.
     When an identifier is found, ExternalTargets is searched first
     which would allow redefining built-in functions by users.}
  //ExternalTargets: TExternalTargetList;
  ExternalTargets: TTargetList;
  {$ENDIF}

implementation

uses
  SysUtils, Math, MiniMath, ParserConsts;

{ TTarget }

procedure TTarget.SetCount(value: integer);
begin
  {$IFDEF DEBUG}
  assert(false, 'Cannot set Count');
  {$ENDIF}
end;

{ TKeyword }

constructor TKeyword.create(anId: integer);
begin
  FCount := anId;
end;

function TKeyword.Calc: float;
begin
  {$IFDEF DEBUG}
  assert(false, 'Cannot calculate an infix operator');
  {$ENDIF}
  result := nan; // removes warnings!
end;

{ TConstant }

constructor TConstant.create(value: Extended);
begin
  inherited Create;
  //FCount := 0;
  FValue := value;
end;

function TConstant.Calc: float;
begin
  result := FValue;
end;

{ TFunction }

constructor TFunction.create(aCount: integer);
begin
  inherited Create;
  FCount := aCount;
end;

{ TVariableFunction }

constructor TVariableFunction.Create(aMinCount: integer);
begin
  inherited Create(0);  //FCount := 0
  {$IFDEF DEBUG}
  assert(aMinCount > 0, 'Minimum number of arguments less than 1');
  {$ENDIF}
  FMinCount := aMinCount;
end;

procedure TVariableFunction.SetCount(value: integer);
begin
  if Count <> value then begin
    {$IFDEF DEBUG}
    assert(value >= MinCount, 'Too few parameters');
    {$ENDIF}
    FCount := value;
  end;
end;

{ TTargetList }

procedure TTargetList.Add(const aName: string; target: TTarget
              {$IFnDEF NO_TARGET_HINT}; const aHint: string = '' {$ENDIF}
              {$IFnDEF NO_NLS2}; const eName: string = '' {$ENDIF});
begin
  {$IFnDEF NO_TARGET_HINT}
  target.Hint := aHint;
  {$ENDIF}
  {$IFnDEF NO_NLS2}
  if ename <> '' then
    target.InternalName := eName
  else
    target.InternalName := aName;
  {$ENDIF}
  {$IFDEF DEBUG}{$IFnDEF NO_SIGIL}
  if FIsInternalList then
    assert(aName[1] <> SSigil, 'Built-in name must not start with "_"')
  {$IFDEF FORCE_SIGIL}
  else
    assert(aName[1] = SSigil, 'External name must start with "_"')
  {$ENDIF};
  {$ENDIF}{$ENDIF}
  try
    FList.AddObject(aName, target);
  except
    target.Free;
    Raise;
  end;
end;

procedure TTargetList.Clear;
var
  i: integer;
begin
  TStringList(FList).BeginUpdate;
  try
    for i := 0 to FList.Count - 1 do
      FList.Objects[i].Free;
    FList.Clear;
  finally
    TStringList(FList).EndUpdate;
  end;
end;

constructor TTargetList.Create;
begin
  inherited;
  FList := TStringList.create;
  TStringList(FList).Duplicates := dupError;
  TStringList(FList).Sorted := true;
end;

function TTargetList.GetCount: integer;
begin
  result := FList.Count;
end;

function TTargetList.Delete(const aName: string): boolean;
var
  i: integer;
begin
  result := TStringList(FList).Find(aName, i);
  if result then begin
    FList.Objects[i].Free;
    FList.Delete(i);
  end;
end;

destructor TTargetList.Destroy;
begin
  Clear;
  FList.free;
  inherited;
end;

function TTargetList.Find(const aName: string; out target: TTarget): boolean;
var
  i: integer;
begin
  result := TStringList(FList).Find(aName, i);
  if result then begin
    target := FList.Objects[i] as TTarget;
  end
  else begin
    target := nil;
  end;
end;

function TTargetList.GetCaseSensitive: boolean;
begin
  result := TStringList(FList).CaseSensitive
end;

function TTargetList.GetTarget(index: integer): TTarget;
begin
  result := FList.Objects[index] as TTarget;
end;

function TTargetList.HasTarget(const aName: string): boolean;
var
  i: integer;
begin
  result := TStringList(FList).Find(aName, i)
end;

{$IFnDEF NO_NLS2}
type
  TCompareProc = function(const s1, s2: string): boolean;

{: Returns the name of the target with an internal name eName. }
function TTargetList.InternalName(const eName: string; out aName: string): boolean;
var
  i: integer;
  comp: TCompareProc;
begin
  if CaseSensitive then
    comp := @AnsiSameStr
  else
    comp := @AnsiSameText;
  for i := 0 to FList.Count-1 do begin
    result := comp(TTarget(FList.Objects[i]).InternalName, eName);
    if result then begin
      aName := FList[i];
      exit;
    end;
  end;
  aName := '';
end;

{$ENDIF}


function TTargetList.NameOf(target: TTarget): string;
var
  i: integer;
begin
  for I := 0 to FList.Count - 1 do
    if FList.Objects[i] as TTarget = target then begin
      result := FList[i];
      exit;
    end;
  result := '';
end;

procedure TTargetList.SetCaseSensitive(value: boolean);
begin
  TStringList(FList).CaseSensitive := value
end;

{trigonometric}

type
  _sin = class(TFunction)
  public
    function Calc: float; override;
  end;

  _cos = class(TFunction)
  public
    function Calc: float; override;
  end;

  _tan = class(TFunction)
  public
    function Calc: float; override;
  end;

  _cot = class(TFunction)
  public
    function Calc: float; override;
  end;

  _csc = class(TFunction)
  public
    function Calc: float; override;
  end;

  _sec = class(TFunction)
  public
    function Calc: float; override;
  end;

  _rad = class(TFunction)
  public
    function Calc: float; override;
  end;

  _deg = class(TFunction)
  public
    function Calc: float; override;
  end;

function _sin.Calc: float;
begin
  result := sin(Args[0]);
end;

function _cos.Calc: float;
begin
  result := cos(Args[0]);
end;

function _tan.Calc: float;
begin
  result := tan(Args[0]);
end;

function _cot.Calc: float;
begin
  result := cotan(Args[0]);
end;

function _csc.Calc: float;
begin
  result := cosecant(Args[0]);
end;

function _sec.Calc: float;
begin
  result := secant(Args[0]);
end;

function _rad.Calc: float;
begin
  result := DegToRad(Args[0]);
end;

function _deg.Calc: float;
begin
  result := RadToDeg(Args[0]);
end;

{inverse trigonometric}

type
  _asin = class(TFunction)
  public
    function Calc: float; override;
  end;

  _acos = class(TFunction)
  public
    function Calc: float; override;
  end;

  _atan = class(TFunction)
  public
    function Calc: float; override;
  end;

  _acot = class(TFunction)
  public
    function Calc: float; override;
  end;

  _acsc = class(TFunction)
  public
    function Calc: float; override;
  end;

  _asec = class(TFunction)
  public
    function Calc: float; override;
  end;

function _asin.Calc: float;
begin
  result := arcsin(Args[0]);
end;

function _acos.Calc: float;
begin
  result := arccos(Args[0]);
end;

function _atan.Calc: float;
begin
  result := arctan(Args[0]);
end;

function _acot.Calc: float;
begin
  result := arctan(1/Args[0]);
end;

function _acsc.Calc: float;
begin
  result := arcsin(1/Args[0]);
end;

function _asec.Calc: float;
begin
  result := arccos(1/Args[0]);
end;

{hyperbolic}

type
  _sinh = class(TFunction)
  public
    function Calc: float; override;
  end;

  _cosh = class(TFunction)
  public
    function Calc: float; override;
  end;

  _tanh = class(TFunction)
  public
    function Calc: float; override;
  end;

  _coth = class(TFunction)
  public
    function Calc: float; override;
  end;

  _csch = class(TFunction)
  public
    function Calc: float; override;
  end;

  _sech = class(TFunction)
  public
    function Calc: float; override;
  end;

function _sinh.Calc: float;
begin
  result := sinh(Args[0]);
end;

function _cosh.Calc: float;
begin
  result := cosh(Args[0]);
end;

function _tanh.Calc: float;
begin
  result := tanh(Args[0]);
end;

function _coth.Calc: float;
begin
  result := coth(Args[0]);
end;

function _csch.Calc: float;
begin
  result := csch(Args[0]);
end;

function _sech.Calc: float;
begin
  result := sech(Args[0]);
end;

{inverse hyperbolic}

type
  _asinh = class(TFunction)
  public
    function Calc: float; override;
  end;

  _acosh = class(TFunction)
  public
    function Calc: float; override;
  end;

  _atanh = class(TFunction)
  public
    function Calc: float; override;
  end;

  _acoth = class(TFunction)
  public
    function Calc: float; override;
  end;

  _acsch = class(TFunction)
  public
    function Calc: float; override;
  end;

  _asech = class(TFunction)
  public
    function Calc: float; override;
  end;

function _asinh.Calc: float;
begin
  result := arcsinh(Args[0]);
end;

function _acosh.Calc: float;
begin
  result := arccosh(Args[0]);
end;

function _atanh.Calc: float;
begin
  result := arctanh(Args[0]);
end;

function _acoth.Calc: float;
begin
  result := arccoth(Args[0]);
end;

{ArcCscH is incorrectly implemented in Math.pas
 included in versions of Delphi prior to Rad Studio 8.0
 (Delphi XE).}

{$IFDEF FIX_ARCCSCH}
  {$IFnDEF USE_MATH_UNIT}
    should not be here!
  {$ENDIF}
function _acsch.Calc: float;
begin
  Result := Ln(Sqrt(1 + (1 / (Args[0] * Args[0]))) + (1 / Args[0]));
end;
{$ELSE}
function _acsch.Calc: float;
begin
  result := arccsch(Args[0])
end;
{$ENDIF}

function _asech.Calc: float;
begin
  result := arcsech(Args[0]);
end;

{general}

type
  _abs = class(TFunction)
  public
    function Calc: float; override;
  end;

  _sqr = class(TFunction)
  public
    function Calc: float; override;
  end;

  _sqrt = class(TFunction)
  public
    function Calc: float; override;
  end;

  _ln = class(TFunction)
  public
    function Calc: float; override;
  end;

  _log = class(TFunction)
  public
    function Calc: float; override;
  end;

  _log10 = class(TFunction)
  public
    function Calc: float; override;
  end;

  _exp = class(TFunction)
  public
    function Calc: float; override;
  end;

  _odd = class(TFunction)
  public
    function Calc: float; override;
  end;

  _floor = class(TFunction)
  public
    function Calc: float; override;
  end;

  _ceil = class(TFunction)
  public
    function Calc: float; override;
  end;

  _int = class(TFunction)
  public
    function Calc: float; override;
  end;

  _sgn = class(TFunction)
  public
    function Calc: float; override;
  end;

  _round = class(TFunction)
  public
    function Calc: float; override;
  end;

  _clamp = class(TFunction)
  public
    function Calc: float; override;
  end;

  _fact = class(TFunction)
  public
    function Calc: float; override;
  end;

  _max = class(TVariableFunction)
  public
    function Calc: float; override;
  end;

  _min = class(TVariableFunction)
  public
    function Calc: float; override;
  end;

  _avg = class(TVariableFunction)
  public
    function Calc: float; override;
  end;

  _ssq = class(TVariableFunction)
  public
    function Calc: float; override;
  end;
  
  _var = class(TVariableFunction)
  public
    function Calc: float; override;
  end;

  _varp = class(TVariableFunction)
  public
    function Calc: float; override;
  end;
  
  _poly = class(TVariableFunction)
  public
    function Calc: float; override;
  end;
 
function _abs.Calc: float;
begin
  result := abs(Args[0]);
end;

function _sqr.Calc: float;
begin
  result := sqr(Args[0]);
end;

function _sqrt.Calc: float;
begin
  result := sqrt(Args[0]);
end;

function _ln.Calc: float;
begin
  result := ln(Args[0]);
end;

function _log.Calc: float;
begin
  result := logn(Args[1], Args[0]);
end;

function _log10.Calc: float;
begin
  result := log10(Args[0]);
end;

function _exp.Calc: float;
begin
  result := exp(Args[0]);
end;

function _odd.Calc: float;
begin
  if odd(trunc(Args[0])) then
    result := 1
  else
    result := 0;
end;

function _floor.Calc: float;
begin
  result := Floor(Args[0]);
end;

function _ceil.Calc: float;
begin
  result := Ceil(Args[0]);
end;

function _int.Calc: float;
begin
  result := trunc(Args[0]);
end;

function _sgn.Calc: float;
begin
  if args[0] > 0 then
    result := 1
  else if args[0] < 0 then
    result := -1
  else
    result := 0;
end;

function _round.Calc: float;
begin
  result := round(Args[0]);
end;

function _clamp.Calc: float;
begin
  result := args[0];
  if args[1] <= args[2] then begin
    if result <  args[1]then
      result := args[1]
    else if result > args[2] then
      result := args[2]
  end;
end;

function _fact.Calc: float;
var
  x: float;
begin
  if args[0] < 0 then
    result := ln(-1)
  else begin
    x := trunc(args[0]);
    result := 1;
    while x > 1 do begin
      result := result * x;
      x := x - 1;
    end;
  end;
end;

{ variable functions }

function _max.Calc: float;
var
  i: integer;
begin
  result := Args[0];
  for i := 1 to Count-1 do
    if Args[i] > result then
      result := Args[i]
end;

function _min.Calc: float;
var
  i: integer;
begin
  result := Args[0];
  for i := 1 to Count-1 do
    if Args[i] < result then
      result := Args[i]
end;

function _avg.Calc: float;
var
  i: integer;
begin
 result := 0;
  for i := 0 to Count-1 do
    result := result + args[i]/count;
end;

function _ssq.Calc: float;
var
  i: integer;
begin
  result := 0;
  for i := 0 to Count-1 do
    result := result + sqr(args[i]);
end;

function _var.Calc: float;
var
  i,n: integer;
  avg: float;
begin
  avg := 0;
  for i := 0 to Count-1 do
    avg := avg + args[i]/count;
  n := pred(count);
  result := 0;
  for i := 0 to Count-1 do
    result := result + (sqr(args[i]-avg))/n;
end;

function _varp.Calc: float;
var
  i: integer;
  avg: float;
begin
  avg := 0;
  for i := 0 to Count-1 do
    avg := avg + args[i]/count;
  result := 0;
  for i := 0 to Count-1 do
    result := result + (sqr(args[i]-avg))/count;
end;

function _poly.Calc: float;
// x, a0, a1, a2
// 0   1   2   3   4
// computes a0 + a1*x + a2*x^2 + a3*x^3... a(count-1)*x^(count-1)
var
  i: integer;
  x: float;
begin
  x := args[0];
  result := 0;
  for i := Count-1 downto 1 do
    result := args[i] + x*result;
end;

procedure LoadInternalTargets;
begin
  InternalTargets.Clear;

  {$IFDEF NO_NLS2}
  InternalTargets.Add(SC_sin, _sin.create(1){$IFnDEF NO_TARGET_HINT}, S_sin_Hint{$ENDIF});
  InternalTargets.Add(SC_cos, _cos.create(1){$IFnDEF NO_TARGET_HINT}, S_cos_Hint{$ENDIF});
  InternalTargets.Add(SC_tan, _tan.create(1){$IFnDEF NO_TARGET_HINT}, S_tan_Hint{$ENDIF});
  InternalTargets.Add(SC_cot, _cot.create(1){$IFnDEF NO_TARGET_HINT}, S_cot_Hint{$ENDIF});
  InternalTargets.Add(SC_csc, _csc.create(1){$IFnDEF NO_TARGET_HINT}, S_csc_Hint{$ENDIF});
  InternalTargets.Add(SC_sec, _sec.create(1){$IFnDEF NO_TARGET_HINT}, S_sec_Hint{$ENDIF});
  InternalTargets.Add(SC_rad, _rad.create(1){$IFnDEF NO_TARGET_HINT}, S_rad_Hint{$ENDIF});
  InternalTargets.Add(SC_deg, _deg.create(1){$IFnDEF NO_TARGET_HINT}, S_deg_Hint{$ENDIF});

  InternalTargets.Add(SC_asin, _asin.create(1){$IFnDEF NO_TARGET_HINT}, S_asin_Hint{$ENDIF});
  InternalTargets.Add(SC_acos, _acos.create(1){$IFnDEF NO_TARGET_HINT}, S_acos_Hint{$ENDIF});
  InternalTargets.Add(SC_atan, _atan.create(1){$IFnDEF NO_TARGET_HINT}, S_atan_Hint{$ENDIF});
  InternalTargets.Add(SC_acot, _acot.create(1){$IFnDEF NO_TARGET_HINT}, S_acot_Hint{$ENDIF});
  InternalTargets.Add(SC_acsc, _acsc.create(1){$IFnDEF NO_TARGET_HINT}, S_acsc_Hint{$ENDIF});
  InternalTargets.Add(SC_asec, _asec.create(1){$IFnDEF NO_TARGET_HINT}, S_asec_Hint{$ENDIF});

  InternalTargets.Add(SC_sinh, _sinh.create(1){$IFnDEF NO_TARGET_HINT}, S_sinh_Hint{$ENDIF});
  InternalTargets.Add(SC_cosh, _cosh.create(1){$IFnDEF NO_TARGET_HINT}, S_cosh_Hint{$ENDIF});
  InternalTargets.Add(SC_tanh, _tanh.create(1){$IFnDEF NO_TARGET_HINT}, S_tanh_Hint{$ENDIF});
  InternalTargets.Add(SC_coth, _coth.create(1){$IFnDEF NO_TARGET_HINT}, S_coth_Hint{$ENDIF});
  InternalTargets.Add(SC_csch, _csch.create(1){$IFnDEF NO_TARGET_HINT}, S_csch_Hint{$ENDIF});
  InternalTargets.Add(SC_sech, _sech.create(1){$IFnDEF NO_TARGET_HINT}, S_sech_Hint{$ENDIF});

  InternalTargets.Add(SC_asinh, _asinh.create(1){$IFnDEF NO_TARGET_HINT}, S_asinh_Hint{$ENDIF});
  InternalTargets.Add(SC_acosh, _acosh.create(1){$IFnDEF NO_TARGET_HINT}, S_acosh_Hint{$ENDIF});
  InternalTargets.Add(SC_atanh, _atanh.create(1){$IFnDEF NO_TARGET_HINT}, S_atanh_Hint{$ENDIF});
  InternalTargets.Add(SC_acoth, _acoth.create(1){$IFnDEF NO_TARGET_HINT}, S_acoth_Hint{$ENDIF});
  InternalTargets.Add(SC_acsch, _acsch.create(1){$IFnDEF NO_TARGET_HINT}, S_acsch_Hint{$ENDIF});
  InternalTargets.Add(SC_asech, _asech.create(1){$IFnDEF NO_TARGET_HINT}, S_asech_Hint{$ENDIF});


  InternalTargets.Add(SC_abs, _abs.create(1){$IFnDEF NO_TARGET_HINT}, S_abs_Hint{$ENDIF});
  InternalTargets.Add(SC_sqr, _sqr.create(1){$IFnDEF NO_TARGET_HINT}, S_sqr_Hint{$ENDIF});
  InternalTargets.Add(SC_sqrt, _sqrt.create(1){$IFnDEF NO_TARGET_HINT}, S_sqrt_Hint{$ENDIF});
  InternalTargets.Add(SC_ln, _ln.create(1){$IFnDEF NO_TARGET_HINT}, S_ln_Hint{$ENDIF});
  InternalTargets.Add(SC_log, _log.create(2){$IFnDEF NO_TARGET_HINT}, S_log_Hint{$ENDIF});
  InternalTargets.Add(SC_log10, _log10.create(1){$IFnDEF NO_TARGET_HINT}, S_log10_Hint{$ENDIF});
  InternalTargets.Add(SC_exp, _exp.create(1){$IFnDEF NO_TARGET_HINT}, S_exp_Hint{$ENDIF});
  InternalTargets.Add(SC_odd, _odd.create(1){$IFnDEF NO_TARGET_HINT}, S_odd_Hint{$ENDIF});
  InternalTargets.Add(SC_floor, _floor.create(1){$IFnDEF NO_TARGET_HINT}, S_floor_Hint{$ENDIF});
  InternalTargets.Add(SC_ceil, _ceil.create(1){$IFnDEF NO_TARGET_HINT}, S_ceil_Hint{$ENDIF});
  InternalTargets.Add(SC_int, _int.create(1){$IFnDEF NO_TARGET_HINT}, S_int_Hint{$ENDIF});
  InternalTargets.Add(SC_sgn, _sgn.create(1){$IFnDEF NO_TARGET_HINT}, S_sgn_Hint{$ENDIF});
  InternalTargets.Add(SC_round, _round.create(1){$IFnDEF NO_TARGET_HINT}, S_round_Hint{$ENDIF});
  InternalTargets.Add(SC_fact, _fact.create(1){$IFnDEF NO_TARGET_HINT}, S_fact_Hint{$ENDIF});
  InternalTargets.Add(SC_clamp, _clamp.create(3){$IFnDEF NO_TARGET_HINT}, S_clamp_Hint{$ENDIF});

  InternalTargets.Add(SC_max, _max.create(1){$IFnDEF NO_TARGET_HINT}, S_max_Hint{$ENDIF});
  InternalTargets.Add(SC_min, _min.create(1){$IFnDEF NO_TARGET_HINT}, S_min_Hint{$ENDIF});
  InternalTargets.Add(SC_avg, _avg.create(1){$IFnDEF NO_TARGET_HINT}, S_avg_Hint{$ENDIF});
  InternalTargets.Add(SC_ssq, _ssq.create(1){$IFnDEF NO_TARGET_HINT}, S_ssq_Hint{$ENDIF});
  InternalTargets.Add(SC_var, _var.create(2){$IFnDEF NO_TARGET_HINT}, S_var_Hint{$ENDIF});
  InternalTargets.Add(SC_varp, _varp.create(1){$IFnDEF NO_TARGET_HINT}, S_varp_Hint{$ENDIF});
  InternalTargets.Add(SC_poly, _poly.create(2){$IFnDEF NO_TARGET_HINT}, S_poly_Hint{$ENDIF});

  InternalTargets.Add(SC_e, TConstant.create(exp(1)){$IFnDEF NO_TARGET_HINT}, S_e_Hint{$ENDIF});
  InternalTargets.Add(SC_pi, TConstant.create(pi){$IFnDEF NO_TARGET_HINT}, S_pi_Hint{$ENDIF});

  InternalTargets.Add(SC_mod, TKeyword.Create(ord(ttPercent)){$IFnDEF NO_TARGET_HINT}, S_mod_Hint{$ENDIF});
  InternalTargets.Add(SC_div, TKeyword.Create(ord(ttColon)){$IFnDEF NO_TARGET_HINT}, S_div_Hint{$ENDIF});
  {$ELSE}
  InternalTargets.Add(SR_sin, _sin.create(1){$IFnDEF NO_TARGET_HINT}, S_sin_Hint{$ENDIF}, SC_sin);
  InternalTargets.Add(SR_cos, _cos.create(1){$IFnDEF NO_TARGET_HINT}, S_cos_Hint{$ENDIF}, SC_cos);
  InternalTargets.Add(SR_tan, _tan.create(1){$IFnDEF NO_TARGET_HINT}, S_tan_Hint{$ENDIF}, SC_tan);
  InternalTargets.Add(SR_cot, _cot.create(1){$IFnDEF NO_TARGET_HINT}, S_cot_Hint{$ENDIF}, SC_cot);
  InternalTargets.Add(SR_csc, _csc.create(1){$IFnDEF NO_TARGET_HINT}, S_csc_Hint{$ENDIF}, SC_csc);
  InternalTargets.Add(SR_sec, _sec.create(1){$IFnDEF NO_TARGET_HINT}, S_sec_Hint{$ENDIF}, SC_sec);
  InternalTargets.Add(SR_rad, _rad.create(1){$IFnDEF NO_TARGET_HINT}, S_rad_Hint{$ENDIF}, SC_rad);
  InternalTargets.Add(SR_deg, _deg.create(1){$IFnDEF NO_TARGET_HINT}, S_deg_Hint{$ENDIF}, SC_deg);

  InternalTargets.Add(SR_asin, _asin.create(1){$IFnDEF NO_TARGET_HINT}, S_asin_Hint{$ENDIF}, SC_asin);
  InternalTargets.Add(SR_acos, _acos.create(1){$IFnDEF NO_TARGET_HINT}, S_acos_Hint{$ENDIF}, SC_acos);
  InternalTargets.Add(SR_atan, _atan.create(1){$IFnDEF NO_TARGET_HINT}, S_atan_Hint{$ENDIF}, SC_atan);
  InternalTargets.Add(SR_acot, _acot.create(1){$IFnDEF NO_TARGET_HINT}, S_acot_Hint{$ENDIF}, SC_acot);
  InternalTargets.Add(SR_acsc, _acsc.create(1){$IFnDEF NO_TARGET_HINT}, S_acsc_Hint{$ENDIF}, SC_acsc);
  InternalTargets.Add(SR_asec, _asec.create(1){$IFnDEF NO_TARGET_HINT}, S_asec_Hint{$ENDIF}, SC_asec);

  InternalTargets.Add(SR_sinh, _sinh.create(1){$IFnDEF NO_TARGET_HINT}, S_sinh_Hint{$ENDIF}, SC_sinh);
  InternalTargets.Add(SR_cosh, _cosh.create(1){$IFnDEF NO_TARGET_HINT}, S_cosh_Hint{$ENDIF}, SC_cosh);
  InternalTargets.Add(SR_tanh, _tanh.create(1){$IFnDEF NO_TARGET_HINT}, S_tanh_Hint{$ENDIF}, SC_tanh);
  InternalTargets.Add(SR_coth, _coth.create(1){$IFnDEF NO_TARGET_HINT}, S_coth_Hint{$ENDIF}, SC_coth);
  InternalTargets.Add(SR_csch, _csch.create(1){$IFnDEF NO_TARGET_HINT}, S_csch_Hint{$ENDIF}, SC_csch);
  InternalTargets.Add(SR_sech, _sech.create(1){$IFnDEF NO_TARGET_HINT}, S_sech_Hint{$ENDIF}, SC_sech);

  InternalTargets.Add(SR_asinh, _asinh.create(1){$IFnDEF NO_TARGET_HINT}, S_asinh_Hint{$ENDIF}, SC_asinh);
  InternalTargets.Add(SR_acosh, _acosh.create(1){$IFnDEF NO_TARGET_HINT}, S_acosh_Hint{$ENDIF}, SC_acosh);
  InternalTargets.Add(SR_atanh, _atanh.create(1){$IFnDEF NO_TARGET_HINT}, S_atanh_Hint{$ENDIF}, SC_atanh);
  InternalTargets.Add(SR_acoth, _acoth.create(1){$IFnDEF NO_TARGET_HINT}, S_acoth_Hint{$ENDIF}, SC_acoth);
  InternalTargets.Add(SR_acsch, _acsch.create(1){$IFnDEF NO_TARGET_HINT}, S_acsch_Hint{$ENDIF}, SC_acsch);
  InternalTargets.Add(SR_asech, _asech.create(1){$IFnDEF NO_TARGET_HINT}, S_asech_Hint{$ENDIF}, SC_asech);

  InternalTargets.Add(SR_abs, _abs.create(1){$IFnDEF NO_TARGET_HINT}, S_abs_Hint{$ENDIF}, SC_abs);
  InternalTargets.Add(SR_sqr, _sqr.create(1){$IFnDEF NO_TARGET_HINT}, S_sqr_Hint{$ENDIF}, SC_sqr);
  InternalTargets.Add(SR_sqrt, _sqrt.create(1){$IFnDEF NO_TARGET_HINT}, S_sqrt_Hint{$ENDIF}, SC_sqrt);
  InternalTargets.Add(SR_ln, _ln.create(1){$IFnDEF NO_TARGET_HINT}, S_ln_Hint{$ENDIF}, SC_ln);
  InternalTargets.Add(SR_log, _log.create(2){$IFnDEF NO_TARGET_HINT}, S_log_Hint{$ENDIF}, SC_log);
  InternalTargets.Add(SR_log10, _log10.create(1){$IFnDEF NO_TARGET_HINT}, S_log10_Hint{$ENDIF}, SC_log10);
  InternalTargets.Add(SR_exp, _exp.create(1){$IFnDEF NO_TARGET_HINT}, S_exp_Hint{$ENDIF}, SC_exp);
  InternalTargets.Add(SR_odd, _odd.create(1){$IFnDEF NO_TARGET_HINT}, S_odd_Hint{$ENDIF}, SC_odd);
  InternalTargets.Add(SR_floor, _floor.create(1){$IFnDEF NO_TARGET_HINT}, S_floor_Hint{$ENDIF}, SC_floor);
  InternalTargets.Add(SR_ceil, _ceil.create(1){$IFnDEF NO_TARGET_HINT}, S_ceil_Hint{$ENDIF}, SC_ceil);
  InternalTargets.Add(SR_int, _int.create(1){$IFnDEF NO_TARGET_HINT}, S_int_Hint{$ENDIF}, SC_int);
  InternalTargets.Add(SR_sgn, _sgn.create(1){$IFnDEF NO_TARGET_HINT}, S_sgn_Hint{$ENDIF}, SC_sgn);
  InternalTargets.Add(SR_round, _round.create(1){$IFnDEF NO_TARGET_HINT}, S_round_Hint{$ENDIF}, SC_round);
  InternalTargets.Add(SR_fact, _fact.create(1){$IFnDEF NO_TARGET_HINT}, S_fact_Hint{$ENDIF}, SC_fact);
  InternalTargets.Add(SR_clamp, _clamp.create(3){$IFnDEF NO_TARGET_HINT}, S_clamp_Hint{$ENDIF}, SC_clamp);

  InternalTargets.Add(SR_max, _max.create(1){$IFnDEF NO_TARGET_HINT}, S_max_Hint{$ENDIF}, SC_max);
  InternalTargets.Add(SR_min, _min.create(1){$IFnDEF NO_TARGET_HINT}, S_min_Hint{$ENDIF}, SC_min);
  InternalTargets.Add(SR_avg, _avg.create(1){$IFnDEF NO_TARGET_HINT}, S_avg_Hint{$ENDIF}, SC_avg);
  InternalTargets.Add(SR_ssq, _ssq.create(1){$IFnDEF NO_TARGET_HINT}, S_ssq_Hint{$ENDIF}, SC_ssq);
  InternalTargets.Add(SR_var, _var.create(2){$IFnDEF NO_TARGET_HINT}, S_var_Hint{$ENDIF}, SC_var);
  InternalTargets.Add(SR_varp, _varp.create(1){$IFnDEF NO_TARGET_HINT}, S_varp_Hint{$ENDIF}, SC_varp);
  InternalTargets.Add(SR_poly, _poly.create(2){$IFnDEF NO_TARGET_HINT}, S_poly_Hint{$ENDIF}, SC_poly);

  InternalTargets.Add(SR_e, TConstant.create(exp(1)){$IFnDEF NO_TARGET_HINT}, S_e_Hint{$ENDIF}, SC_e);
  InternalTargets.Add(SR_pi, TConstant.create(pi){$IFnDEF NO_TARGET_HINT}, S_pi_Hint{$ENDIF}, SC_pi);

  InternalTargets.Add(SR_mod, TKeyword.Create(ord(ttPercent)){$IFnDEF NO_TARGET_HINT}, S_mod_Hint{$ENDIF}, SC_mod);
  InternalTargets.Add(SR_div, TKeyword.Create(ord(ttColon)){$IFnDEF NO_TARGET_HINT}, S_div_Hint{$ENDIF}, SC_div);
  {$ENDIF}
end;

initialization
  InternalTargets := TTargetList.create;
  {$IFDEF DEBUG}{$IFnDEF NO_SIGIL}
  InternalTargets.FIsInternalList := true;
  {$ENDIF}{$ENDIF}
  LoadInternalTargets;
  {$IFnDEF NO_EXTERNALS}
  ExternalTargets := TTargetList.Create;
  {$ENDIF}
finalization
  {$IFnDEF NO_EXTERNALS}
  ExternalTargets.Free;
  {$ENDIF}
  InternalTargets.Free;
end.
